#:include 'common.fypp'
module auxil
  use constants, only: sp, dp, pi
  implicit none

  private
#:for prec in PRECS
  public :: zolo_coeff_${prec}$
#:endfor

#:for prec in PRECS
  interface zolo_coeff_${prec}$
    module procedure zolo_coeff_1_${prec}$
    module procedure zolo_coeff_r_${prec}$
  end interface
#:endfor

  contains

#:for prec in PRECS
    !> Computes the coefficients c_j and a_j
    subroutine zolo_coeff_r_${prec}$(con, tol, rk, cj, aj)
      real(${prec}$), intent(in) :: con, tol
      integer, intent(in) :: rk
      real(${prec}$), intent(out) :: cj(2*rk), aj(rk)

      integer :: i, rk2, j
      real(${prec}$) :: alpha, tmp1, kk, sn, cn

      alpha = acos(1._${prec}$ / con)
      call mellipke_${prec}$(alpha, tol, kk, tmp1)
      rk2 = 2 * rk

      do i = 1, rk2
        tmp1 = real(i, kind=${prec}$) * kk / (real(rk2, kind=${prec}$) + 1._${prec}$)
        call mellipj_${prec}$(tmp1, alpha, tol, sn, cn)
        cj(i) = sn**2 / cn**2
      end do

      do i = 1, rk
        aj(i) = 1._${prec}$
        do j = 1, rk
          aj(i) = aj(i) * (cj(2*i-1) - cj(2*j))
          if (i /= j) then
            aj(i) = aj(i) / (cj(2*i-1) - cj(2*j-1))
          end if
        end do
        aj(i) = -aj(i)
     end do
    end subroutine

    !> Computes the coefficients c_j and a_j for r=1
    subroutine zolo_coeff_1_${prec}$(con, tol,  cj, aj)
      real(${prec}$), intent(in) :: con, tol
      real(${prec}$), intent(out) :: cj(2), aj

      integer :: i
      real(${prec}$) :: alpha, tmp1, kk, sn, cn

      alpha = acos(1._${prec}$ / con)
      call mellipke_${prec}$(alpha, tol, kk, tmp1)

      do i = 1, 2
        tmp1 = real(i, kind=${prec}$) * kk / 3._${prec}$
        call mellipj_${prec}$(tmp1, alpha, tol, sn, cn)
        cj(i) = sn**2 / cn**2
      end do

      aj = cj(2) - cj(1)
    end subroutine

    !> Complete elliptic integrals of first and second kind
    subroutine mellipke_${prec}$(alpha, tol, k, e)
      real(${prec}$), intent(in) :: alpha, tol
      real(${prec}$), intent(out) :: k, e

      real(${prec}$) :: mm, a0, b0, s0, a1, b1, w1, c1
      integer :: i1

      a0 = 1._${prec}$
      b0 = cos(alpha)
      s0 = sin(alpha)**2
      i1 = 0
      mm = 1._${prec}$

      do while (mm > tol)
        a1 = 0.5_${prec}$ * (a0 + b0)
        b1 = sqrt(a0 * b0)
        c1 = 0.5_${prec}$ * (a0 - b0)
        i1 = i1 + 1
        w1 = (2._${prec}$**i1) * (c1**2)
        mm = w1
        s0 = s0 + w1
        a0 = a1
        b0 = b1
      end do

      k = 0.5_${prec}$ * real(pi, kind=${prec}$) / a1
      e = k * (1._${prec}$ - 0.5_${prec}$ * s0)
    end subroutine

    !> Jacobi elliptic functions
    subroutine mellipj_${prec}$(u, alpha, tol, sn, cn)
      real(${prec}$), intent(in) :: u, alpha, tol
      real(${prec}$), intent(out) :: sn, cn

      integer, parameter :: mchunk=1000
      real(${prec}$) :: a(mchunk), b(mchunk), c(mchunk), phin(mchunk)
      integer  :: nin, i

      a(1) = 1._${prec}$
      b(1) = cos(alpha)
      c(1) = sin(alpha)

      i = 1
      nin = 0
      do while (abs(c(i)) > tol .and. i < mchunk)
        i = i + 1
        a(i) = (a(i-1) + b(i-1)) * 0.5_${prec}$
        b(i) = sqrt(a(i-1) * b(i-1))
        c(i) = (a(i-1) - b(i-1)) * 0.5_${prec}$

        if (abs(c(i)) <= tol .and. abs(c(i-1)) > tol) then
          nin = i - 1
        end if
      end do

      phin(1) = u
      phin(i) = 2._${prec}$**nin * a(i) * u

      do while (i > 1)
        i = i - 1
        phin(i) = phin(i+1)
        if (nin >= i) then
          phin(i) = 0.5_${prec}$ * (asin(c(i+1) * sin(phin(i+1)) / a(i+1)) + phin(i+1))
        end if
      end do

      sn = sin(phin(1))
      cn = cos(phin(1))
    end subroutine
#:endfor

end module
